@@ -6,7 +6,7 @@ use warnings FATAL => 'all';
 use parent qw(Exporter);
 use Carp qw(croak);
 
-our @EXPORT_OK = qw(diff dselect dsplit patch);
+our @EXPORT_OK = qw(diff dselect dsplit dtraverse patch);
 
 sub _validate_meta($) {
     my $d = shift;
@@ -24,11 +24,11 @@ Struct::Diff - Recursive diff tools for nested perl structures
 
 =head1 VERSION
 
-Version 0.58
+Version 0.61
 
 =cut
 
-our $VERSION = '0.58';
+our $VERSION = '0.61';
 
 =head1 SYNOPSIS
 
@@ -134,9 +134,10 @@ sub diff($$;@) {
             $d->{'N'} = $b;
         }
     } elsif ((ref $a eq 'ARRAY') and ($a ne $b)) {
+        my $s; # statuses collector
         for (my $i = 0; $i < @{$a} and $i < @{$b}; $i++) {
             my $tmp = diff($a->[$i], $b->[$i], %opts);
-            if (keys %{$tmp}) {
+            if (map { $s->{$_} = 1 } keys %{$tmp}) { # true if keys exists
                 $tmp->{'I'} = $i if ($hidden);
                 push @{$d->{'D'}}, $tmp;
             } else {
@@ -147,6 +148,7 @@ sub diff($$;@) {
             if ($opts{'noR'}) {
                 $hidden = 1;
             } else {
+                $s->{'R'} = 1;
                 map { push @{$d->{'D'}}, { 'R' => $opts{'trimR'} ? undef : $_ } } @{$a}[@{$b}..$#{$a}];
             }
         }
@@ -154,17 +156,14 @@ sub diff($$;@) {
             if ($opts{'noA'}) {
                 $hidden = 1;
             } else {
+                $s->{'A'} = 1;
                 map { push @{$d->{'D'}}, { 'A' => $_ } } @{$b}[@{$a}..$#{$b}];
             }
         }
 
-        my $s = { map { $_, 1 } map { keys %{$_} } exists $d->{'D'} ? @{$d->{'D'}} : { 'U' => 1 } };
-        delete $s->{'I'}; # ignored -- not a status
-
-        if (keys %{$s} == 1 and not ($hidden or exists $s->{'D'})) { # all has same status - drop D and return as is
-            my $n = (keys(%{$s}))[0];
-            map { $_ = $_->{$n} } @{$d->{'D'}};
-            $d->{$n} = delete $d->{'D'};
+        if ((my @k = keys %{$s}) == 1 and not ($hidden or exists $s->{'D'})) { # all have same status - return it
+            map { $_ = $_->{$k[0]} } @{$d->{'D'}};
+            $d->{$k[0]} = delete $d->{'D'};
         }
     } elsif ((ref $a eq 'HASH') and ($a ne $b)) {
         for my $key (keys %{{ %{$a}, %{$b} }}) { # go througth united uniq keys
@@ -199,11 +198,9 @@ sub diff($$;@) {
                 delete $d->{$s};
             }
         }
-    } else { # treat others as scalars
-        unless ((not defined $a and not defined $b) or ((defined $a and defined $b) and ($a eq $b))) {
-            $d->{'O'} = $a unless ($opts{'noO'});
-            $d->{'N'} = $b unless ($opts{'noN'});
-        }
+    } elsif (not(defined $a and defined $b and $a eq $b or not defined $a and not defined $b)) { # other types
+        $d->{'O'} = $a unless ($opts{'noO'});
+        $d->{'N'} = $b unless ($opts{'noN'});
     }
     $d->{'U'} = $a unless ($hidden or $opts{'noU'} or keys %{$d});
 
@@ -312,6 +309,50 @@ sub dsplit($) {
     return $s;
 }
 
+=head2 dtraverse
+
+Traverse through diff invoking callback functions for subdiff statuses. Important: path (secont argument,
+passed to callback functions) is actual for callback lifetime and will be changed afterwards.
+
+    my $opts = {
+        A => sub { print "added:", $_[0], "depth:", @{$_[1]} },
+        U => sub { print "unchaanged: ", $_[0] },
+    };
+    dtraverse($diff, $opts);
+
+=cut
+
+sub dtraverse($$;$);
+sub dtraverse($$;$) {
+    my ($d, $o, $p) = (shift, shift, shift || []);
+    _validate_meta($d);
+
+    if (exists $d->{'D'}) {
+        if (ref $d->{'D'} eq 'ARRAY') {
+            for (my $i = 0; $i < @{$d->{'D'}}; $i++) {
+                push @{$p}, [$i];
+                dtraverse($d->{'D'}->[$i], $o, $p);
+                pop @{$p};
+            }
+        } else { # HASH
+            for my $k (keys %{$d->{'D'}}) {
+                push @{$p}, { 'keys' => [$k] };
+                dtraverse($d->{'D'}->{$k}, $o, $p);
+                pop @{$p};
+            }
+        }
+    } elsif (exists $d->{'U'} and $o->{'U'}) {
+        $o->{'U'}($d->{'U'}, $p);
+    } elsif (exists $d->{'A'} and $o->{'A'}) {
+        $o->{'A'}($d->{'A'}, $p);
+    } elsif (exists $d->{'R'} and $o->{'R'}) {
+        $o->{'R'}($d->{'R'}, $p);
+    } else {
+        $o->{'N'}($d->{'N'}, $p) if (exists $d->{'N'} and $o->{'N'});
+        $o->{'O'}($d->{'O'}, $p) if (exists $d->{'O'} and $o->{'O'});
+    }
+}
+
 =head2 patch
 
 Apply diff
@@ -325,37 +366,31 @@ sub patch($$) {
     my ($s, $d) = @_;
     _validate_meta($d);
 
-    ${$s} = $d->{'N'} if (exists $d->{'N'});
-
     if (exists $d->{'D'}) {
         if (ref $d->{'D'} eq 'ARRAY') {
             for my $i (0..$#{$d->{'D'}}) {
-                next if (exists $d->{'D'}->[$i]->{'U'});
                 my $si = exists $d->{'D'}->[$i]->{'I'} ? $d->{'D'}->[$i]->{'I'} : $i; # use provided index
                 if (exists $d->{'D'}->[$i]->{'D'} or exists $d->{'D'}->[$i]->{'N'}) {
                     patch(ref $s->[$si] ? $s->[$si] : \$s->[$si], $d->{'D'}->[$i]);
-                    next;
-                }
-                if (exists $d->{'D'}->[$i]->{'A'}) {
+                } elsif (exists $d->{'D'}->[$i]->{'A'}) {
                     push @{$s}, $d->{'D'}->[$i]->{'A'};
-                    next;
+                } elsif (exists $d->{'D'}->[$i]->{'R'}) {
+                    pop @{$s};
                 }
-                pop @{$s} if (exists $d->{'D'}->[$i]->{'R'});
             }
         } else { # HASH
             for my $k (keys %{$d->{'D'}}) {
-                next if (exists $d->{'D'}->{$k}->{'U'});
                 if (exists $d->{'D'}->{$k}->{'D'} or exists $d->{'D'}->{$k}->{'N'}) {
                     patch(ref $s->{$k} ? $s->{$k} : \$s->{$k}, $d->{'D'}->{$k});
-                    next;
-                }
-                if (exists $d->{'D'}->{$k}->{'A'}) {
+                } elsif (exists $d->{'D'}->{$k}->{'A'}) {
                     $s->{$k} = $d->{'D'}->{$k}->{'A'};
-                    next;
+                } elsif (exists $d->{'D'}->{$k}->{'R'}) {
+                    delete $s->{$k};
                 }
-                delete $s->{$k} if (exists $d->{'D'}->{$k}->{'R'});
             }
         }
+    } elsif (exists $d->{'N'}) {
+        ${$s} = $d->{'N'};
     }
 
     return 1;
@@ -411,7 +446,7 @@ L<http://search.cpan.org/dist/Struct-Diff/>
 
 =head1 SEE ALSO
 
-L<Data::Diff>, L<Data::Difference>
+L<Data::Diff>, L<Data::Difference>, L<Data::Deep>
 
 L<Array::Diff>, L<Array::Compare>, L<Algorithm::Diff>, L<Data::Compare>, L<Hash::Diff>, L<Test::Struct>,
 L<Struct::Compare>
